@@ -14,7 +14,7 @@
  * The following % expansions are supported:
  *
  * 	%b	<blk>			block number
- * 	%B	<blkcount>		interpret blkcount as blkcount
+ * 	%B	<blkcount>		integer
  * 	%c	<blk2>			block number
  * 	%Di	<dirent>->ino		inode number
  * 	%Dn	<dirent>->name		string
@@ -46,7 +46,6 @@
  * 	%q	ext2fs_get_pathname of directory <dir>
  * 	%Q	ext2fs_get_pathname of directory <ino> with <dir> as
  * 			the containing directory.
- * 	%r	<blkcount>		interpret blkcount as refcount
  * 	%s	<str>			miscellaneous string
  * 	%S	backup superblock
  * 	%X	<num> hexadecimal format
@@ -258,7 +257,7 @@ static _INLINE_ void expand_at_expression(e2fsck_t ctx, char ch,
 /*
  * This function expands '%IX' expressions
  */
-static _INLINE_ void expand_inode_expression(ext2_filsys fs, char ch,
+static _INLINE_ void expand_inode_expression(char ch,
 					     struct problem_context *ctx)
 {
 	struct ext2_inode	*inode;
@@ -292,8 +291,7 @@ static _INLINE_ void expand_inode_expression(ext2_filsys fs, char ch,
 		printf("%u", large_inode->i_extra_isize);
 		break;
 	case 'b':
-		if (fs->super->s_feature_ro_compat &
-		    EXT4_FEATURE_RO_COMPAT_HUGE_FILE) 
+		if (inode->i_flags & EXT4_HUGE_FILE_FL)
 			printf("%llu", inode->i_blocks +
 			       (((long long) inode->osd2.linux2.l_i_blocks_hi)
 				<< 32));
@@ -397,11 +395,9 @@ static _INLINE_ void expand_dirent_expression(ext2_filsys fs, char ch,
 }
 
 static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
-					       int *first,
 					       struct problem_context *ctx)
 {
 	e2fsck_t e2fsck_ctx = fs ? (e2fsck_t) fs->priv_data : NULL;
-	const char *m;
 
 	if (!ctx)
 		goto no_context;
@@ -418,26 +414,11 @@ static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
 #endif
 		break;
 	case 'B':
-		if (ctx->blkcount == BLOCK_COUNT_IND)
-			m = _("indirect block");
-		else if (ctx->blkcount == BLOCK_COUNT_DIND)
-			m = _("double indirect block");
-		else if (ctx->blkcount == BLOCK_COUNT_TIND)
-			m = _("triple indirect block");
-		else if (ctx->blkcount == BLOCK_COUNT_TRANSLATOR)
-			m = _("translator block");
-		else
-			m = _("block #");
-		if (*first && islower(m[0]))
-			fputc(toupper(*m++), stdout);
-		fputs(m, stdout);
-		if (ctx->blkcount >= 0) {
 #ifdef EXT2_NO_64_TYPE
-			printf("%d", ctx->blkcount);
+		printf("%d", ctx->blkcount);
 #else
-			printf("%lld", (long long) ctx->blkcount);
+		printf("%lld", (long long)ctx->blkcount);
 #endif
-		}
 		break;
 	case 'c':
 #ifdef EXT2_NO_64_TYPE
@@ -481,13 +462,6 @@ static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
 	case 'Q':
 		print_pathname(fs, ctx->dir, ctx->ino);
 		break;
-	case 'r':
-#ifdef EXT2_NO_64_TYPE
-		printf("%d", ctx->blkcount);
-#else
-		printf("%lld", (long long) ctx->blkcount);
-#endif
-		break;
 	case 'S':
 		printf("%u", get_backup_sb(NULL, fs, NULL, NULL));
 		break;
@@ -529,13 +503,13 @@ void print_e2fsck_message(e2fsck_t ctx, const char *msg,
 			expand_at_expression(ctx, *cp, pctx, &first, recurse);
 		} else if (cp[0] == '%' && cp[1] == 'I') {
 			cp += 2;
-			expand_inode_expression(fs, *cp, pctx);
+			expand_inode_expression(*cp, pctx);
 		} else if (cp[0] == '%' && cp[1] == 'D') {
 			cp += 2;
 			expand_dirent_expression(fs, *cp, pctx);
 		} else if ((cp[0] == '%')) {
 			cp++;
-			expand_percent_expression(fs, *cp, &first, pctx);
+			expand_percent_expression(fs, *cp, pctx);
 		} else {
 			for (i=0; cp[i]; i++)
 				if ((cp[i] == '@') || cp[i] == '%')
