@@ -36,10 +36,10 @@ extern void com_err_va (const char *whoami, errcode_t code, const char *fmt,
 
 extern char const *error_message (long);
 extern void (*com_err_hook) (const char *, long, const char *, va_list);
-extern void (*set_com_err_hook (void (*) (const char *, long, 
+extern void (*set_com_err_hook (void (*) (const char *, long,
 					  const char *, va_list)))
 	(const char *, long, const char *, va_list);
-extern void (*reset_com_err_hook (void)) (const char *, long, 
+extern void (*reset_com_err_hook (void)) (const char *, long,
 					  const char *, va_list);
 extern int init_error_table(const char * const *msgs, long base, int count);
 
@@ -49,12 +49,16 @@ extern void add_to_error_table(struct et_list *new_table);
 
 /* Provided for Heimdall compatibility */
 extern const char *com_right(struct et_list *list, long code);
-extern void initialize_error_table_r(struct et_list **list, 
-				     const char **messages, 
+extern void initialize_error_table_r(struct et_list **list,
+				     const char **messages,
 				     int num_errors,
 				     long base);
 extern void free_error_table(struct et_list *et);
 
+/* Provided for compatibility with other com_err libraries */
+extern int et_list_lock(void);
+extern int et_list_unlock(void);
+
 #define __COM_ERR_H
 #define __COM_ERR_H__
 #endif /* !defined(__COM_ERR_H) && !defined(__COM_ERR_H__)*/
